unit Just3D;

interface
{-$DEFINE SCANLINE} { removed under Delphi2 }
{-$DEFINE BLUR}

uses
{$ifdef win32}
 windows,classes,graphics;
{$else}
 Xlib,Qt,classes,QGraphics;
{$endif}

type
 Float=single;

procedure Init3D(width,height:integer);
Function  LoadBMP256(FileName:string):integer;
Function SetTexture(index:integer):boolean;

Procedure ClearPixels;
procedure BeginPoly(x,y:integer; z:float; u,v:integer; s:float);
procedure ScanPoly(x2,y2:integer; iz2:float; iu2,iv2:integer; is2:float);
procedure EndPoly;
procedure SetPixels(Canvas:TCanvas);

implementation

{$IFDEF VER90}{$UNDEF  SCANLINE}{$ENDIF} // no ScanLine under Delphi 2
{$IFDEF LINUX}{$DEFINE SCANLINE}{$ENDIF} // XPutImage doesn't work yet :((
{$ifdef linux}{$ifndef scanline}{$define convert}{$endif}{$endif}

procedure Done3D(ClearTexture:boolean); forward;

Const
 ShadeCount=63;

type
 TBMP256=packed record
  bfType:array[1..2] of char;
  bfSize:LongInt;
  brReserved:LongInt;
  bgOffBits:LongInt;
  biSize:LongInt;
  biWidth:LongInt;
  biHeight:LongInt;
  biPlanes:word;
  biBitCount:word;
  biCompression:longint;
  biSizeImage:longint;
  biXPelsPerMeter:longint;
  biYPelsPerMeter:longint;
  biClrUsed:Longint;
  biClrImportant:longint;
  WinPal:array[0..255] of integer;
 end;

 PTexture=^TTextureBits;
 TTexture=record
  width :integer;
  height:integer;
  shades:array[0..ShadeCount,0..255] of integer;
 end;
 TTextureBits=record
  width :integer;
  height:integer;
  shades:array[0..ShadeCount,0..255] of integer;
  bits  :array[word] of byte;
 end;

 PWords=^TWords;
 TWords=array[word] of Word;
 PSmallInts=^TSmallInts;
 TSmallInts=array[word] of SmallInt;
 PIntegers=^TIntegers;
 TIntegers=array[word] of Integer;
 PFloats=^TFloats;
 TFloats=array[word] of Float;

var
 Textures:TStringList;

 maxx,maxy:integer;

 XLeft,XRight:PSmallInts;

 rSLeft,rSRight:PFloats;
 rULeft,rURight:PFloats;
 rVLeft,rVRight:PFloats;
 rZLeft,rZRight:PFloats;

 Top,Bottom:integer;
 x0,y0:integer;
    z0:float;
 u0,v0:integer;
    s0:float;

{$ifdef SCANLINE}
 Bitmap:TBitmap;
{$else}
 Pixels:PIntegers;
{$ifdef win32}
 BitmapInfo2:TBitmapInfoHeader;
 BitmapInfo :PBitmapInfo;
{$else}
 Display:PDisplay;
 Screen :integer;
 Window :Xlib.Window;
 Visual :PVisual;
 Context:TGC;
 Depth  :integer;
 Bitmap :PXImage;
{$endif}
{$endif}

procedure Init3D(width,height:integer);
 begin
  Done3D(False);
  maxx:=width;
  maxy:=height;
  getmem(xleft,maxy*SizeOf(SmallInt));
  getmem(xright,maxy*SizeOf(SmallInt));
  getmem(rsleft,maxy*SizeOf(float));
  getmem(rsright,maxy*SizeOf(float));
  getmem(ruleft,maxy*SizeOf(float));
  getmem(ruright,maxy*SizeOf(float));
  getmem(rvleft,maxy*SizeOf(float));
  getmem(rvright,maxy*SizeOf(float));
  getmem(rzleft,maxy*SizeOf(float));
  getmem(rzright,maxy*SizeOf(float));
 {$IFDEF SCANLINE}
  Bitmap:=TBitmap.Create;
  Bitmap.Width:=Width;
  Bitmap.Height:=Height;
 // under Kylix BETA4, you can not set PF before Size...
  Bitmap.PixelFormat:=pf32bit;
 {$ELSE}
  getmem(pixels,maxx*maxy*sizeof(integer));
 {$ifdef win32}
  With BitmapInfo2 do begin
   biWidth :=maxx;
   biHeight:=-maxy;
  end;
 {$else}
  Display:=QTDisplay;
  Screen :=XDefaultScreen(Display);
  Window :=XRootWindow(Display,Screen);
  Visual :=XDefaultVisual(Display,Screen);
  Context:=XDefaultGC(Display,Screen);
  Depth  :=XDisplayPlanes(Display,Screen);
  WriteLn('Creating a ',Depth,'bits Bitmap...');
  Bitmap :=XCreateImage(Display,Visual,Depth,ZPixmap,0,PChar(Pixels),maxx,maxy,Depth,0);
 {$endif}
 {$ENDIF}
 end;

function Shade(color,level:integer):integer;
 var
  r,g,b:integer;
  a:real;
 begin
  a:=level/ShadeCount;
  r:=round(a*((color) and $FF));
  g:=round(a*((color shr 8) and $FF));
  b:=round(a*((color shr 16) and $FF));
 {$ifdef convert}
  WriteLn('Computing colors for ',Depth,'bits');
  case Depth of
   15: result:=(r shr 3)+(g shr 3) shl 5+(b shr 3) shl 10; // 0rrrrrgg.gggbbbbb
   16: result:=(r shr 3)+(g shr 2) shl 5+(b shr 3) shl 11; // rrrrrggg.gggbbbbb
  else result:=r+g shl 8+b shl 16;
  end;
 {$else}
  result:=r+g shl 8+b shl 16;
 {$endif}
 end;

function LoadBMP256(FileName:string):integer;
 var
  header:TBMP256;
  fs:TFileStream;
  tx:PTexture;
  i,j:integer;
 begin
  result:=Textures.IndexOf(FileName);
  if result>=0 then exit;
  fs:=TFileStream.Create(FileName,0);
  fs.ReadBuffer(header,SizeOf(header));
  GetMem(tx,SizeOf(TTexture)+header.biWidth*header.biHeight);
  tx.Width:=header.biWidth;
  tx.Height:=header.biHeight;
  for i:=0 to ShadeCount do begin
   for j:=0 to 255 do begin
    tx.shades[i,j]:=Shade(header.WinPal[j],i);
   end;
  end;
// fs.ReadBuffer(tx.Bits,tx.width*tx.height);
  j:=tx.Width*tx.Height;
  for i:=0 to tx.Height-1 do begin
   dec(j,tx.Width);
   fs.ReadBuffer(tx.Bits[j],tx.Width);
  end;
  fs.Free;
  Result:=Textures.AddObject(FileName,TObject(tx));
 end;

var
 t:PTexture;

function SetTexture(index:integer):boolean;
 begin
  result:=index<Textures.Count;
  if result then t:=PTexture(Textures.Objects[index]);
 end;

procedure ClearPixels;
{$IFDEF BLUR}
 var
  x,y:integer;
  c:^integer;
 begin
 {$ifndef scanline}c:=pointer(Pixels);{$endif}
  for y:=0 to maxy-1 do begin
  {$ifdef scanline}c:=Bitmap.ScanLine[y];{$endif}
   for x:=0 to maxx-1 do begin
    c^:=(c^ and ($FFFFFF-$010101)) shr 1;
    inc(c);
   end;
  end;
{$ELSE}
 begin
 {$ifdef scanline}
  with Bitmap.Canvas do begin
   Brush.Color:=clBlack;
   FillRect(Rect(0,0,maxx,maxy));
  end;
 {$ifdef linux}Bitmap.FreePixmap;{$endif} // probably a little bug with BETA4
 {$else}
  FillChar(Pixels^,maxx*maxy*sizeof(integer),0);
 {$endif}
{$ENDIF}
 end;

procedure BeginPoly(x,y:integer; z:float; u,v:integer; s:float);
 begin
  FillChar(XLeft^,maxy*SizeOf(SmallInt),3);
  FillChar(XRight^,maxy*SizeOf(SmallInt),255);
  Top:=MaxY;
  Bottom:=0;
  x0:=x;
  y0:=y;
  z0:=z;
  u0:=u;
  v0:=v;
  s0:=abs(s*ShadeCount);
 end;

procedure delta(var d,i:integer);
 begin
  if d<0 then begin d:=-d; i:=-1 end else i:=+1;
 end;

procedure ScanPoly(x2,y2:integer; iz2:float; iu2,iv2:integer; is2:float);
 var
  x1,y1:integer;
  dx,dy:integer;
  ix,iy:integer;
  ii:integer;

  z1,u1,v1,s1:real;
  z2,u2,v2,s2:real;
  dz,du,dv,ds:real;
  dd:real;
  xy:real;

  procedure range;
   begin
    if (y1>=0)and(y1<MaxY) then begin

     if (x1<XLeft[y1]) then begin
      if y1<Top then Top:=y1;
      if y1>Bottom then Bottom:=y1;
       XLeft[y1]:=x1;
      rZLeft[y1]:=z1;
      rULeft[y1]:=u1;
      rVLeft[y1]:=v1;
      rSLeft[y1]:=s1;
     end;

     if (x1>XRight[y1]) then begin
      if y1<Top then Top:=y1;
      if y1>Bottom then Bottom:=y1;
       XRight[y1]:=x1;
      rZRight[y1]:=z1;
      rURight[y1]:=u1;
      rVRight[y1]:=v1;
      rSRight[y1]:=s1;
     end;
    end;
   end;


 begin
  x1:=x0; x0:=x2;
  y1:=y0; y0:=y2;
  dx:=x2-x1; delta(dx,ix);
  dy:=y2-y1; delta(dy,iy);

  if z0=0 then z1:=1 else z1:=1/z0;
  z0:=iz2;
  if z0=0 then z2:=1 else z2:=1/iz2;
  u1:=u0*z1; u0:=iu2; u2:=iu2*z2;
  v1:=v0*z1; v0:=iv2; v2:=iv2*z2;
  is2:=abs(is2*ShadeCount);
  s1:=s0;    s0:=is2; s2:=is2;

  if dx>dy then begin

   dd:=1/dx;
   dz:=(z2-z1)*dd;
   du:=(u2-u1)*dd;
   dv:=(v2-v1)*dd;
   ds:=(s2-s1)*dd;
   xy:=y1;
   dd:=(y2-y1)*dd;

   for ii:=0 to dx do begin
    range;
    inc(x1,ix);
    xy := xy + dd;
    y1 :=round(xy);
    z1 := z1 + dz;
    u1 := u1 + du;
    v1 := v1 + dv;
    s1 := s1 + ds;
   end;

  end else begin

   if dy=0 then exit;

   dd:=1/dy;
   dz:=(z2-z1)*dd;
   du:=(u2-u1)*dd;
   dv:=(v2-v1)*dd;
   ds:=(s2-s1)*dd;
   xy:=x1;
   dd:=(x2-x1)*dd;

   for ii:=0 to dy do begin
    range;
    xy := xy + dd;
    x1 :=round(xy);
    inc(y1,iy);
    z1 := z1 + dz;
    u1 := u1 + du;
    v1 := v1 + dv;
    s1 := s1 + ds;
   end;
  end;
 end;

procedure PolyLine15_16(x1,x2,y:integer; Var Pixels);
 var
  dd:single;
  z1,u1,v1,s1:single;
  dz,du,dv,ds:single;
  dx:integer;
  line:PWords;
 begin
  if x1>=MaxX then exit;
  dd:=1/(x2-x1);
  u1:=rULeft[y]; du:=(rURight[y]-u1)*dd;
  v1:=rVLeft[y]; dv:=(rVRight[y]-v1)*dd;
  z1:=rZLeft[y]; dz:=(rZRight[y]-z1)*dd;
  s1:=rSLeft[y]; ds:=(rSRight[y]-s1)*dd;
  if x2>=MaxX then x2:=MaxX-1;
  line:=Addr(Pixels);
  line:=Addr(line(.maxx*y.));
  for dx:=x1 to x2 do begin
   if (dx>=0) then line^[dx]:=t.shades[round(s1),t.bits[t.width*round(v1/z1)+round(u1/z1)]];
   u1:=u1+du;
   v1:=v1+dv;
   z1:=z1+dz;
   s1:=s1+ds;
  end;
 end;

procedure PolyLine(x1,x2,y:integer; line:PIntegers);
 var
  dd:single;
  z1,u1,v1,s1:single;
  dz,du,dv,ds:single;
  dx:integer;
 begin
  if x1>=MaxX then exit;
  dd:=1/(x2-x1);
  u1:=rULeft[y]; du:=(rURight[y]-u1)*dd;
  v1:=rVLeft[y]; dv:=(rVRight[y]-v1)*dd;
  z1:=rZLeft[y]; dz:=(rZRight[y]-z1)*dd;
  s1:=rSLeft[y]; ds:=(rSRight[y]-s1)*dd;
  if x2>=MaxX then x2:=MaxX-1;
  for dx:=x1 to x2 do begin
   if (dx>=0) then line^[dx]:=t.shades[round(s1),t.bits[t.width*round(v1/z1)+round(u1/z1)]];
   u1:=u1+du;
   v1:=v1+dv;
   z1:=z1+dz;
   s1:=s1+ds;
  end;
 end;

procedure EndPoly;
 var
  y,x1,x2:integer;
 begin
  if top<0 then top:=0;
  if Bottom>=MaxY then Bottom:=MaxY-1;
  for y:=Top to Bottom do begin
   x1:=XLeft[y];
   x2:=XRight[y];
   if x2>x1 then
   {$ifdef convert}
    case Depth of
     15,16:PolyLine15_16(x1,x2,y,Pixels);
//     24:PolyLine24(x1,x2,y,Pixels);
     32:Polyline(x1,x2,y,@Pixels[maxx*y]);
    end;
   {$else}
    Polyline(x1,x2,y,{$IFDEF SCANLINE}Bitmap.ScanLine[y]{$ELSE}@Pixels[maxx*y]{$ENDIF});
   {$endif}
  end;
 end;

procedure SetPixels(Canvas:TCanvas);
 const
  i:byte=0;
 begin
{$IFDEF SCANLINE}
  Canvas.Draw(0,0,Bitmap);
{$ELSE}
 {$ifdef win32}
  SetDiBitsToDevice(Canvas.handle,0,0,maxx-1,maxy-1,0,0,0,maxy-1,Pixels,BitmapInfo^,0);
 {$else}
  XPutImage(Display,Window,Context,Bitmap,0,0,0,0,maxx,maxy);
 {$endif}
{$ENDIF}
 end;

procedure Done3D(ClearTexture:boolean);
 var
  i:integer;
 begin
  if ClearTexture then begin
   for i:=0 to Textures.Count-1 do freemem(PTexture(Textures.Objects[i]));
   Textures.Clear;
  end;
  FreeMem(xleft);
  FreeMem(xright);
  FreeMem(rsleft);
  FreeMem(rsright);
  FreeMem(ruleft);
  FreeMem(ruright);
  FreeMem(rvleft);
  FreeMem(rvright);
  FreeMem(rzleft);
  FreeMem(rzright);
 {$ifdef SCANLINE}
  Bitmap.Free;
 {$else}
 {$ifdef linux}
  if Bitmap<>nil then begin
   XDestroyImage(Bitmap);
   Bitmap:=nil;
  end;
 {$endif}
  FreeMem(Pixels);
 {$endif}
 end;

initialization
 Textures:=TStringList.Create;
{$IFNDEF SCANLINE}
{$IFDEF WIN32}
 BitmapInfo:=@BitmapInfo2;
 With BitmapInfo2 do begin
  biSize         :=SizeOf(TBitmapInfoHeader);
  biplanes       :=1;
  bibitcount     :=32; { for True-Color }
  bicompression  :=BI_RGB;
  biSizeImage    :=0;
  bixpelspermeter:=0;
  biypelspermeter:=0;
  biclrused      :=0;
  biclrimportant :=0;
 end;
{$ENDIF}
{$ENDIF}

finalization
 Done3D(true);
 Textures.Free;

end.
