unit MD2MAP;

{ this unit can compile under both Delphi & Kylix...
  BUT I use a Windows File Mapping where I should use a Memory or FileStream :)
  ...BTW is it possible to map files under Linux ?
}

interface

uses
{$ifdef win32}
 windows,
{$else}
 classes,
{$endif}
 sysutils;

Type
 THeader=record
  Ident      :integer;
  Version    :integer;

  SkinWidth  :integer;
  SkinHeight :integer;
  FrameSize  :integer;

  SkinCount  :integer;
  VertexCount:integer;
  MapCount   :integer;
  FaceCount  :integer;
  GLCmdCount :integer;
  FrameCount :integer;

  SkinOfs    :integer;
  MapOfs     :integer;
  FaceOfs    :integer;
  FrameOfs   :integer;
  GLCmdOfs   :integer;
  EndOfFile  :integer;
 end;

 TSkinName =array[0..63] of char;
 TSkinNames=array[0..31] of TSkinName;

 TMap=record u,v:smallint end;
 TMaps=array[0..2047] of TMap;

 TFace=record
  a,b,c:smallint;
  ta,tb,tc:smallint;
 end;
 TFaces=array[0..4095] of TFace;

 TVertex=record
  x,y,z:byte;
  light:byte;
 end;
 TFrame=record
  Scale:record x,y,z:single end;
  Trans:record x,y,z:single end;
  Name :array[0..15] of char;
  Verts:array[0..2047] of TVertex;
 end;

 TMD2MAP=class
  fName:string;
 {$ifdef win32}
  hFile:THandle;
  hMap :THandle;
 {$else}
  Stream:TMemoryStream;
 {$endif}
  hPtr :pointer;
  header:^THeader;
  Skins :^TSkinNames;
  Maps  :^TMaps;
  Faces :^TFaces;
  Frames:^TFrame;
 public
  Constructor Create(AFileName:string);
  Destructor Destroy; override;
  Procedure GetPoint(frame,index:integer; var x,y,z:single);
  function FrameName(frame:integer):string;
 end;

implementation

Constructor TMD2MAP.Create(AFileName:string);
 begin
  fName:=AFileName;
 {$ifdef win32}
  hFile:=CreateFile(PChar(fName),GENERIC_READ,FILE_SHARE_READ,nil,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,0);
  if hFile=INVALID_HANDLE_VALUE then raise Exception.Create('Unable to open file '+fName);
  hMap :=CreateFileMapping(hFile,nil,PAGE_READONLY,0,0,nil);
  if hMap=0 then raise exception.Create('Unable to Map file '+fName);
  hPtr:=MapViewOfFile(hMap,FILE_MAP_READ,0,0,0);
  if hPtr=nil then raise exception.Create('Unable to view file '+fName);
 {$else}
  Stream:=TMemoryStream.Create;
  Stream.LoadFromFile(AFileName);
  hPtr:=Stream.Memory;
 {$endif}
  header:=hPtr;
  Skins :=pointer(integer(hPtr)+header.skinofs);
  Maps  :=pointer(integer(hPtr)+header.MapOfs);
  Faces :=pointer(integer(hPtr)+header.FaceOfs);
  Frames:=pointer(integer(hPtr)+header.FrameOfs);
 end;

Destructor TMD2MAP.Destroy;
 begin
 {$ifdef win32}
  UnMapViewOfFile(hPtr);
  CloseHandle(hMap);
  CloseHandle(hFile);
 {$else}
  Stream.Free;
 {$endif}
 end;

Procedure TMD2MAP.GetPoint(frame,index:integer; var x,y,z:single);
 var
  F:^TFrame;
 begin
  F:=pointer(integer(Frames)+frame*Header.FrameSize);
  with F^ do begin
   x:=Verts[index].x*Scale.x+Trans.x;
   y:=Verts[index].y*Scale.y+Trans.y;
   z:=Verts[index].z*Scale.z+Trans.z;
  end;
 end;

function TMD2MAP.FrameName(frame:integer):string;
 var
  F:^TFrame;
 begin
  F:=pointer(integer(Frames)+frame*Header.FrameSize);
  Result:=F^.Name;
 end;


end.
