unit MD2a;

{ an half hour to port this pure Delphi 2 application to Kylix Beta 4. }

{ BTW : my Kylix installation
  PIII550 64Mo
  HDD 800Mo + 13Mo of SWAP
  Mandrake 7.0 (minimal developpemnt installation with KDE1.1.2)
  rpm -U glibc-2.1.3-16mdkB.i586.rpm
  sh setup.sh -m
 (need to add -m to force installation)
}

interface

uses
{$ifdef win32} // Delphi Units
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ShellAPI,StdCtrls, ExtCtrls,
{$else} // Kylix Units
  SysUtils, Types, Classes, QGraphics, QControls, QForms, QDialogs,
  QExtCtrls,
{$endif}
  MD2MAP, Scenes,Just3D;

type
  TForm1 = class(TForm)
    Timer1: TTimer;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormPaint(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure FormMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure FormMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure FormMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure Timer1Timer(Sender: TObject);
  private
  {$ifdef win32}
   // Windows specific Drag&Drop !
    procedure WMDropFiles(Var Msg:TMessage); message WM_DROPFILES;
  {$endif}
  public
    MD2:TMD2MAP;
    SCN:TScene;
    PLY:TPlayer;
    mx,my:integer;
    mm:boolean;
    frame:integer;
    procedure SetFileName(FileName:string);
  end;

var
  Form1: TForm1;

implementation

{$ifdef win32}
{$R *.DFM}
{$else}
{$R *.xfm} // Kylix use "xfm" files where Delphi use "DFM" files
{$endif}

procedure TForm1.FormCreate(Sender: TObject);
begin
{$ifdef win32} // Windows Drag&Drop
 DragAcceptFiles(Handle,True);
{$endif}
 SCN:=TScene.Create(Self,0,0,0);
// This line for Kylix only, cause it call FormResize while SCN=nil !
// Under windows, FormResize is probably called by Windows AFTER FormCreate !
 SCN.Resize(ClientWidth,ClientHeight);
 SetFileName('TRIS.MD2');
 SetFileName('Athena.bmp');
end;

procedure TForm1.FormResize(Sender: TObject);
begin
// This test for Kylix only, under windows SCN is never nil !
 if Assigned(SCN) then
 SCN.Resize(ClientWidth,ClientHeight);
end;

{$ifdef win32} // Drag&Drop
procedure TForm1.WMDropFiles(Var Msg:TMessage);
Const
 MaxLen=512;
var
 count:integer;
 s:string;
 i:integer;
 l:integer;
 f:TForm1;
begin
 count:=DragQueryFile(Msg.wParam,uint(-1),nil,0);
 for i:=0 to count-1 do begin
  SetLength(s,MaxLen);
  l:=DragQueryFile(Msg.wParam,i,pchar(s),length(s));
  if i=0 then f:=Self else begin
   f:=TForm1.Create(Application);
   f.SetBounds(Left+21,Top+21,Width,Height);
   f.Show;
  end;
  f.SetFileName(Copy(s,1,l));
 end;
 DragFinish(Msg.wParam);
end;
{$endif}

procedure TForm1.FormDestroy(Sender: TObject);
begin
{$ifdef win32} // Drag&Drop
 DragAcceptFiles(Handle,False);
{$endif}
 MD2.Free;
 PLY.Free;
 SCN.Free;
end;

procedure TForm1.SetFileName(FileName:string);
var
 i:integer;
 f:TFace;
begin
 if Uppercase(ExtractFileExt(FileName))='.BMP' then begin
  PLY.Texture:=LoadBMP256(FileName);
  exit;
 end;
 MD2.Free;
 PLY.Free;
 MD2:=TMD2MAP.Create(FileName);
 Caption:=ExtractFileName(FileName)+' - '
         +IntToStr(MD2.Header.VertexCount)+' vertices, '
         +IntToStr(MD2.Header.FaceCount)+' faces, '
         +IntToStr(MD2.Header.FrameCount)+' frames';
 PLY:=TPlayer.Create(SCN);
 integer(PLY.Color):=clSilver;
 PLY.Define(MD2.Header.VertexCount);
 frame:=0;
 for i:=0 to PLY.VertexCount-1 do with PLY.Vertices[i] do MD2.GetPoint(frame,i,x,y,z);
 for i:=0 to MD2.Header.FaceCount-1 do begin
  with MD2.Faces[i] do begin
   f:=PLY.AddFace(a,b,c);
   f.u1:=MD2.Maps[ta].u;
   f.v1:=MD2.Maps[ta].v;
   f.u2:=MD2.Maps[tb].u;
   f.v2:=MD2.Maps[tb].v;
   f.u3:=MD2.Maps[tc].u;
   f.v3:=MD2.Maps[tc].v;
  end;
 end;
 SCN.ComputeNormal;
 SCN.Angle.x:=-90;  // valeurs empiriques
 SCN.Angle.y:=  0;
 SCN.Angle.z:=  0;
 PLY.Zoom:=1024;
end;

procedure TForm1.FormPaint(Sender: TObject);
begin
 SCN.Render;
end;

procedure TForm1.FormMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
 mx:=x;
 my:=y;
 mm:=true;
 SetCaptureControl(self);
end;

procedure TForm1.FormMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
begin
 if mm then begin
  if ssRight in Shift then begin
   PLY.Zoom:=PLY.Zoom+my-y;
   SCN.Angle.z:=SCN.Angle.z-(mx-x);
  end else begin
   SCN.Angle.x:=SCN.Angle.x-(my-y);
   SCN.Angle.y:=SCN.Angle.y+(mx-x);
  end;
  mx:=x;
  my:=y;
  Paint;
 end;
end;

procedure TForm1.FormMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
 SetCaptureControl(nil);
 mm:=false;
end;

procedure TForm1.Timer1Timer(Sender: TObject);
var
 i:integer;
begin
 if MD2=nil then exit;
 inc(frame);
 if frame=MD2.Header.FrameCount then frame:=0;
 for i:=0 to PLY.VertexCount-1 do with PLY.Vertices[i] do MD2.GetPoint(frame,i,x,y,z);
 SCN.ComputeNormal;
 SCN.Render;
end;

end.
