unit Scenes;

{ to compile this under Kylix, I'd just to had the RGB function...}

interface
uses
 Just3D,
{$ifdef win32}
 windows,classes,forms,graphics;
{$else}
 classes,qforms,qgraphics;
{$endif}

Type
  float=single;

  TScene=class; // forward

  TPoint=record
   x,y,z:float;
  end;

  TColor=packed record
   R,G,B,A:byte
  end;

  PVertex=^TVertex;
  TVertex=record
   X,Y,Z:float;    // 3D position
   rx,ry,rz:float; // Rotated position
   nx,ny,nz:float; // Normal vector
   gx,gy,gz:float; // Gouraud Vector (Normal rotation)
   sx,sy:integer;  // Screen position
   sz:float;       // Depth
  end;
  TVertices=array[word] of TVertex;
  PVertices=^TVertices;

  TFace =class
   A,B,C:PVertex;  // 3 Vertices
   u1,v1:integer;  // Texture mapping
   u2,v2:integer;  // Texture mapping
   u3,v3:integer;  // Texture mapping
   Color:TColor;
   nx,ny,nz:float; // Normal vector
   rx,ry,rz:float; // Rotated Normal
   zz:float;       //
   function GotPoint(P:PVertex):boolean;
   procedure ComputeZ;
   procedure ComputeNormal;
   procedure Draw;
  end;

  TObject3D=class
   Scene      :TScene;
   VertexCount:integer;
   Vertices   :PVertices;
   Faces      :TList;
   Color      :TColor;
   Zoom       :float;
   texture    :integer;
   Constructor Create(AScene:TScene);
   Destructor Destroy; override;
   Procedure Clear;
   function AddFace(a,b,c:integer):TFace;
   procedure ComputeNormal;
  end;

  TLamp=class
   Scene :TScene;
   Center:TPoint;
   Color :TColor;
  end;

  TScene=class
   Eye     :TPoint; // Point of View
   Center  :TPoint;
   Angle   :TPoint; // X,Y,Z rotation
   Color   :TColor; // Background color
   Faces   :TList;  // Global Face List
   Lamps   :TList;  // Light Sources
   Objects :TList;  // 3D Objects
   Target  :TForm;
  public
   procedure Project;
   Procedure Resize(Width,Height:integer);
   Constructor Create(ATarget:TForm; ox,oy,oz:float);
   Destructor Destroy; override;
   procedure Clear;
   procedure ComputeNormal;
   procedure Render;
  end;

 //-------------------------------------------------------------------------//
  TSphere=class(TObject3D)
   Procedure Define(Step,Radius1,Radius2,Radius3:integer);
  end;

  TPlayer=class(TObject3D)
   Procedure Define(AVertexCount:integer);
  end;

implementation

Constructor TScene.Create(ATarget:TForm; ox,oy,oz:float);
 begin
  Eye.x:=ox;
  Eye.y:=oy;
  Eye.z:=oz;
  Angle.X:=0;
  Angle.Y:=0;
  Angle.Z:=0;
  Faces:=TList.Create;
  Objects:=TList.Create;
  Lamps:=TList.Create;
  Target:=ATarget;
 end;

Destructor TScene.Destroy;
 begin
  Clear;
  Faces.Free;
  Objects.Free;
  Lamps.Free;
  inherited;
 end;

procedure TScene.Clear;
 begin
  While Lamps.Count>0 do TLamp(Lamps.Last).Free;
  While Objects.Count>0 do TObject3D(Objects.Last).Free;
  Faces.Clear;
 end;

procedure TScene.ComputeNormal;
 var
  i:integer;
 begin
  for i:=0 to Objects.Count-1 do TObject3D(Objects[i]).ComputeNormal;
 end;

Procedure TScene.Resize(Width,Height:integer);
 begin
  Center.Z:=150;
  Center.X:=Width/2;
  Center.Y:=Height/2;
  Init3D(Width,Height);
  Render;
 end;

Procedure TScene.Project;
 var
  cosx,cosy,cosz,sinx,siny,sinz:float;
  i,j:integer;
  tx,ty,tz:float;
 begin
  cosx:=-cos(Angle.X*PI/180);
  cosy:=+cos(Angle.Y*PI/180);
  cosz:=+cos(angle.Z*PI/180);
  sinx:=-sin(Angle.X*PI/180);
  siny:=+sin(Angle.Y*PI/180);
  sinz:=+sin(Angle.Z*PI/180);
  for i:=0 to Objects.Count-1 do
   with TObject3D(Objects[i]) do begin
    for j:=0 to Faces.Count-1 do begin
     with TFace(Faces[j]) do begin
     { rotate x }
      ty:=(ny*cosx)-(nz*sinx);
      tz:=(ny*sinx)+(nz*cosx);
     { rotate y }
      tx:=(nx*cosy)+(tz*siny);
      rz:=(tz*cosy)-(nx*siny);
     { rotate z }
      rx:=(tx*cosz)-(ty*sinz);
      ry:=(tx*sinz)+(ty*cosz);
     end;
    end;
    for j:=0 to VertexCount-1 do begin
     with Vertices^[j] do begin
     // gouraud
     { rotate x }
      ty:=(ny*cosx)-(nz*sinx);
      tz:=(ny*sinx)+(nz*cosx);
     { rotate y }
      tx:=(nx*cosy)+(tz*siny);
      gz:=(tz*cosy)-(nx*siny);
     { rotate z }
      gx:=(tx*cosz)-(ty*sinz);
      gy:=(tx*sinz)+(ty*cosz);
     // points
     { rotate x }
      ty:=( y*cosx)-( z*sinx);
      tz:=( y*sinx)+( z*cosx);
     { rotate y }
      tx:=( x*cosy)+(tz*siny);
      rz:=(tz*cosy)-( x*siny);
     { rotate z }
      rx:=(tx*cosz)-(ty*sinz);
      ry:=(tx*sinz)+(ty*cosz);
     { project }
      sz:=rz+Center.Z; if sz=0 then sz:=0.001;
      sx:=Round(Center.X+Zoom*rx/sz);
      sy:=Round(Center.Y+Zoom*ry/sz);
     end;
    end;
   end;
 end;

function ZSort(f1,f2:pointer):integer; far;
 var
  z:float;
 begin
  z:=TFace(f2).zz-TFace(f1).zz;
  if z<-0.1 then result:=-1 else
  if z>+0.1 then result:=+1 else result:=0;
 end;

Procedure TScene.Render;
 var
  i:integer;
 begin
  if Objects.Count=0 then exit;
  if not SetTexture(TObject3D(Objects.Last).Texture) then exit;
  ClearPixels;
  Project;
  for i:=0 to Faces.Count-1 do TFace(Faces[i]).ComputeZ;
  Faces.Sort(ZSort);
  for i:=0 to Faces.Count-1 do TFace(Faces[i]).Draw;
  SetPixels(Target.Canvas);
 end;

Constructor TObject3D.Create(AScene:TScene);
 begin
  Scene:=AScene;
  Scene.Objects.Add(Self);
  VertexCount:=0;
  Vertices:=nil;
  Faces:=TList.Create;
  Zoom:=1;
 end;

Destructor TObject3D.Destroy;
 begin
  Scene.Objects.Remove(Self);
  Clear;
  Faces.Free;
  Inherited;
 end;

Procedure TObject3D.Clear;
 var
  i:integer;
 begin
  for i:=0 to Faces.Count-1 do Scene.Faces.Remove(Faces[i]);
  FreeMem(Vertices);
  VertexCount:=0;
  Faces.Clear;
 end;

function TObject3D.AddFace(a,b,c:integer):TFace;
 begin
  Result:=TFace.Create;
  Result.A:=@Vertices^[a];
  Result.B:=@Vertices^[b];
  Result.C:=@Vertices^[c];
  Result.Color:=Color;
  Faces.Add(Result);
  Scene.Faces.Add(Result);
 end;

procedure TObject3D.ComputeNormal;
 var
  i,j:integer;
  v:PVertex;
  f:TFace;
  x,y,z,c:float;
 begin
  for i:=0 to Faces.Count-1 do TFace(Faces[i]).ComputeNormal;
  for i:=0 to VertexCount-1 do begin
   x:=0;
   y:=0;
   z:=0;
   c:=0;
   v:=@Vertices[i];
   for j:=0 to Faces.Count-1 do begin
    f:=Faces[j];
    if (f.a=v)or(f.b=v)or(f.c=v) then begin
     x:=x+f.nx;
     y:=y+f.ny;
     z:=z+f.nz;
     c:=c+1;
    end;
   end;
   x:=x/c;
   y:=y/c;
   z:=z/c;
   c:=sqrt(x*x+y*y+z*z);
   if c=0 then c:=1;
   v.nx:=x/c;
   v.ny:=y/c;
   v.nz:=z/c;
  end;
 end;

function TFace.GotPoint(P:PVertex):boolean;
 begin
  Result:=(A=P)or(B=P)or(C=P);
 end;

procedure TFace.ComputeZ;
 begin
  zz:=(a.sz+b.sz+c.sz)/3;
 end;

procedure TFace.ComputeNormal;
 var
  x1,y1,z1,x2,y2,z2,vl:float;
 begin
  x1:=b.x-a.x;
  y1:=b.y-a.y;
  z1:=b.z-a.z;
  x2:=c.x-a.x;
  y2:=c.y-a.y;
  z2:=c.z-a.z;
  nx:=y1*z2-y2*z1;
  ny:=z1*x2-z2*x1;
  nz:=x1*y2-x2*y1;
  vl:=sqrt(nx*nx+ny*ny+nz*nz);
  if vl<>0 then begin
   nx:=nx/vl;
   ny:=ny/vl;
   nz:=nz/vl;
  end;
 end;

Procedure TFace.Draw;
 begin
  if rz>=0 then begin
   BeginPoly(a.sx,a.sy,a.sz,u1,v1,a.gz);
   ScanPoly (b.sx,b.sy,b.sz,u2,v2,b.gz);
   ScanPoly (c.sx,c.sy,c.sz,u3,v3,c.gz);
   ScanPoly (a.sx,a.sy,a.sz,u1,v1,a.gz);
   EndPoly;
  end;
 end;

Procedure TSphere.Define(Step,Radius1,Radius2,Radius3:integer);
 var
  a1:Float;
  vi:integer;
  s1:integer;
  s2:integer;
  z2:float;
 begin
  Clear;

  VertexCount:=Step*Step+Step;
  Getmem(Vertices,VertexCount*SizeOf(TVertex));

  a1:=PI/Step;
  vi:=0;
  for s1:=0 to Step do begin
   z2:=cos(s1*a1)*Radius3;
   for s2:=0 to Step-1 do begin
    with Vertices^[vi] do begin
     x:=(cos(s2*2*a1)*sin(s1*a1)*Radius1);
     y:=(sin(s2*2*a1)*sin(s1*a1)*Radius2);
     z:=z2;
    end;
    inc(vi);
   end;
  end;

  for s1:=0 to Step do begin
   for s2:=0 to Step-1 do begin
    AddFace(s1*Step+s2,s1*Step+(s2+1) mod Step,(s1*Step+s2+Step) mod vi);
    AddFace(s1*Step+(s2+1) mod Step,(s1*Step+(s2+1) mod Step + Step) mod vi,(s1*Step+ s2+Step) mod vi);
   end;
  end;

 end;

procedure TPlayer.Define(AVertexCount:integer);
 begin
  VertexCount:=AVertexCount;
  GetMem(Vertices,VertexCount*SizeOf(TVertex));
 end;

end.
